// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __emitc
#define __emitc

#include "hashset.h"
#include "str.h"
#include "grammar.h"
#include "nfa.h"

void LIBGLR_EXPORT EmitBuiltinSkeletonFile(outbuf *dst);

// --- everyting below this line needs to dissapear from the headers

class SkeletonFile;
class FrozenDfaTransition;

class EmitGlobals
{
public:
    EmitGlobals();
    Grammar *m_g;
    Production* m_p;
    CodeSnip* m_snip;
    LRStateSet* m_state;
    LRAction* m_action;
    LRActionRange *m_actionrange;
    int m_action_range_index;
    int m_action_range_size;
    OrderedTokenSet* m_tokenset;
    TokenSetAction* m_tokensetaction;
    int m_lexstate;
    FrozenDfaTransition* m_range;
    Symbol* m_s;
    int m_sindex;
    String m_ctype;
    int m_max_production_symbols;
    SkeletonFile *m_sk;
    int m_skipindex;
    UIntSet *m_compressed_range;
    UIntSetRange *m_action_range_item;
    int m_compressed_break;
    compressed_lxitem *m_compressed_lxitem;
};

enum stype { NULL_STMT, IF_STMT, FOREACH_STMT, SUBST_STMT, TXT_STMT };

class Statement
{
    int evalloops(EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
public:
    stype m_stype;
    String m_label, m_txt;
    int m_tok;
    vector<Statement*> *m_strue, *m_selseifs, *m_selse, *m_sloop;
    Statement();
    ~Statement();
    int eval(EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
};

Statement *if_stmt(const char *label, vector<Statement*> *strue, vector<Statement*> *selseifs, vector<Statement*> *selse);
Statement *for_each_stmt(const char *label, vector<Statement*> *sloop);
Statement *subst_stmt(int tok);
Statement *txt_stmt(const char *txt);

class SkeletonFile
{
public:
    hashmap<String,vector<Statement*>*> m_sections;
    Grammar *m_g;

    SkeletonFile(Grammar *g);
    ~SkeletonFile();
    void clear();
    bool parse(inbuf *sk, outbuf *err, outbuf *trace);
    int emit(String sectionname, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
};

int emitC(Grammar *g, emitoptions &options);

extern String FriendlyCType(String s);
extern String SymbolFriendlyName(Symbol *s);
extern String ProductionFriendlyName(Production *p);
extern int Emit_Reduce_Src(const char *src, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
extern int Emit_Src(const char *src, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
extern int DoNT(const char *sec, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
extern int DoSYM(const char *sec, const char *sym, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);
extern int Parse_SkeletonFile(inbuf *buf, outbuf *obuf, outbuf *trace, SkeletonFile *skfile);
extern int evaluate(vector<Statement*> *stmts, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace);

#endif /* __emitc */
