// Copyright (c) 2008-2015 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "grammar.h"
#include "rx.h"
#ifdef _WIN32
#include <windows.h>
#endif
#include "chtest.h"
#include "tokensetmerge.h"
#include "emitC.h"
#include <string.h>
#include "version.h"

bool emitgrammar(Grammar *g, emitoptions &options)
{
    if( g->m_errcount )
        return false;
    outbuf *log = 0;
    if( g->m_logfile.length() )
        log = new_file_outbuf(g->m_logfile.c_str());

    bool bExit = false;

    if( ! g->Solve(log) )
    {
        if( g->m_errcount && g->m_err )
            ob_printf(g->m_err, "%d error(s)\n", g->m_errcount);
        bExit = true;
    }
    if( log )
    {
        free_ob(log);
        log = 0;
    }

    if( g->m_errcount )
        bExit = true;

    if( bExit )
        return false;

    return g->Emit(options) ? true : false;
}

int main(int argc, char *argv[])
{
    bool bShowHelp = false;
    String infile;
    String outfile;
    String logfile;
    String skeletonfile;
    bool bLogDfa = false;
    GrammarSolution RequestedGrammarSolution = DeterministicParser;
    String EmitLanguage = "C";
    String name = "parser";
    outbuf *verbose = 0;
    outbuf *err = new_h_outbuf("<stdout>", H_STDERR, 0, 1);
    bool bLogParse = false;
    bool bLogFirsts = false;
    bool bLogClosure = false;
    bool bLogFollows = false;
    hashset<String> defines;

	String arg;
	wchar_t *warg = 0;
	size_t warglen = 0;
    defines.insert("usepoundline");

    for( int i = 1; i < argc; ++i )
    {
		const char *arg = argv[i];
        if( strcmp(arg, "--nopoundline") == 0 )
        {
            defines.erase("usepoundline");
            continue;
        }
        if( strcmp(arg, "--license") == 0 )
        {
            fputs(  "Permission is hereby granted, free of charge, to any person obtaining a copy\n"
                    "of this software and associated documentation files (the \"Software\"), to deal\n"
                    "in the Software without restriction, including without limitation the rights\n"
                    "to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n"
                    "copies of the Software, and to permit persons to whom the Software is\n"
                    "furnished to do so, subject to the following conditions:\n"
                    "\n"
                    "The above copyright notice and this permission notice shall be included in\n"
                    "all copies or substantial portions of the Software.\n"
                    "\n"
                    "THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
                    "IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
                    "FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n"
                    "AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n"
                    "LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n"
                    "OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\n"
                    "THE SOFTWARE.\n", stdout);
            return 0;
        }
        if( strcmp(arg, "--help") == 0 )
        {
            bShowHelp = true;
            break;
        }
        if( strcmp(arg, "--dfa") == 0 )
        {
            bLogDfa = true;
            continue;
        }

        if( arg[0] == '-' )
        {
            switch( arg[1] )
            {
            case 'k':
                skeletonfile = arg+2;
                break;
            case 'K':
                {
                outbuf *obstdout = new_h_outbuf("<stdout>", H_STDOUT, 0, 1);
                EmitBuiltinSkeletonFile(obstdout);
                free_ob(obstdout);
                exit(1);
                }
                break;
            case 'd':
            case 'D':
                defines.insert(arg+2);
                break;
            case 'o':
                outfile = arg+2;
                break;
            case 'l':
                logfile = arg+2;
                break;
            case 's':
                {
                String solution = arg+2;
                if( solution == "best" )
                    RequestedGrammarSolution = UseBestAchievableSolution;
                else if( solution == "lr" )
                    RequestedGrammarSolution = DeterministicParser;
                else if( solution == "glr" )
                    RequestedGrammarSolution = NondeterministicParser;
                else if( solution == "sglr" )
                    RequestedGrammarSolution = NondeterministicParser_ScannerLess;
                }
                break;
            case 'L':
                EmitLanguage = arg+2;
                if( EmitLanguage == "cpp" || EmitLanguage == "CPP" || EmitLanguage == "c" )
                    EmitLanguage = "C";
                {
                    char buf[1024];
                    sprintf(buf, "emit_language_%s", EmitLanguage.c_str());
                    defines.insert(buf);
                }
                break;
            case 'n':
                name = arg+2;
                break;
            case 'v':
                verbose = new_h_outbuf("<stdout>", H_STDOUT, 0, 1);
                break;
            case 'f':
                {
                    
                    for( int i = 2; arg[i]; ++i )
                    {
                        switch(arg[i])
                        {
                        case 'p':
                            bLogParse = true;
                            break;
                        case 'f':
                            bLogFirsts = true;
                            break;
                        case 'c':
                            bLogClosure = true;
                            break;
                        case 'F':
                            bLogFollows = true;
                            break;
                        default:
                            break;
                        }
                    }
                }
                break;
            default:
                ob_printf(err, "Skipping unknown option \"%s\".\n", arg);
                break;
            }
        }
        else
        {
            infile = arg;
        }
    }

    if( bShowHelp )
    {
        fputs("glrgen input -ooutput [--help] [-llog] [-s(lr|glr|sglr|best)] [-L(cpp|py)] [-f[p][f][c][F]]\n"
                                "\n"
                                "glrgen " GLRGEN_VERSION "\n"
                                "Copyright (c) 2008-2015 Mighty Heave Software LLC.  http://www.mightyheave.com\n"
                                "use --license to see the license\n"
                                "\n"
                                "input      : a glrgen input file\n"
                                "-o         : the output file\n"
                                "--help     : print this message and exit\n"
                                "--nopoundline : do not add the #line compiler directives to the output (C/C++ only)\n"
                                "-l         : the log file.  Optional, but trust me, you'll want this.\n"
                                "-v         : be verbose, output to stdout.\n"
                                "-n         : the grammar name.  (default is \"parser\") Declare the function\n"
                                "             extern grammar_interface* Get_{Name}_GrammarInterface()\n"
                                "             in order to use the grammar.\n"
                                "-s         : the desired generated solution source, if you're confused\n"
                                "             at the options, just use 'best'.\n"
                                "             Note, not all solutions are available for all languages.\n"
                                "             * 'lr' requires both deterministic scanner and deterministic\n"
                                "                parser (no shift/reduce conflicts) or an error will be\n"
                                "                emitted.\n"
                                "             * 'glr' requires a deterministic scanner, but not a deterministic\n"
                                "               parser (shift/reduce conflicts ok).  If the scanner is\n"
                                "               nondeterminstic, an error will be emitted.\n"
                                "             * 'sglr' requires neither a deterministic grammar nor a\n"
                                "               deterministic scanner.\n"
                                "             * 'best' [default] picks the most deterministic solution available\n"
                                "               given the supplied grammar.\n"
                                "-L         : the output language\n"
                                "             * 'cpp' [default] emits a C++ parser (sglr,glr,lr).\n"
                                "             * 'py' emits a python parser (lr).\n"
                                "--dfa      : also log the resulting nfa/dfa lexers.\n"
                                "-f[p][f][c][F] : also log algorithms as they run.\n"
                                "                 p log the parse\n"
                                "                 f log the firsts\n"
                                "                 c log the state closure\n"
                                "                 F log the follows\n"
                                "-K         : emit the built-in skeleton file to stdout and exit.\n"
                                "-k{skfile} : use skfile instead of the built-in skeleton file.\n", stdout
            );
        return 0;
    }

    Grammar *g = new Grammar();
    g->m_bLogDfa = bLogDfa;
    g->m_RequestedGrammarSolution = RequestedGrammarSolution;
    g->m_verbose = verbose;
    g->m_err = err;
    g->m_bLogParse = bLogParse;
    g->m_bLogFirsts = bLogFirsts;
    g->m_bLogClosure = bLogClosure;
    g->m_bLogFollows = bLogFollows;
    g->m_name = name;
    g->m_EmitLanguage = EmitLanguage;

    if( ! g->Parse(infile.c_str()) )
    {
        if( err )
            ob_printf(err, "%d Errors.\n", g->m_errcount);
        exit(1);
    }

    g->m_fname_root = outfile;
    g->m_logfile = logfile;
    g->m_outfile = outfile;
    emitoptions options;
    for( hashset<String>::iterator cur = defines.begin(), end = defines.end(); cur != end; ++cur )
        options.defined_symbols.define(cur->c_str());
    options.m_skeletonfile = skeletonfile;
    emitgrammar(g, options);
    if( g->m_errcount > 0 )
    {
        if( err )
            ob_printf(err, "%d Errors.\n", g->m_errcount);
        exit(1);
    }
    delete g;
    return 0;
}
