// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __nfa_h
#define __nfa_h
#include "glrexport.h"
#include "str.h"
#include "intset.h"
#include "c_all_internal.h"

#define STATE_MARKER_STARTSTATE   (1)
#define STATE_MARKER_ENDSTATE     (2)
#define STATE_MARKER_STARTCAPTURE (4)
#define STATE_MARKER_ENDCAPTURE   (8)

class Nfa;
class Rx;

class LIBGLR_EXPORT FrozenDfaTransition
{
public:
    FrozenDfaTransition();
    FrozenDfaTransition(int start, int stop, int to);
    int m_start, m_stop, m_to;
};

// A smaller dfa, faster and more convenient
class LIBGLR_EXPORT FrozenDfa
{
public:
    FrozenDfa();
    ~FrozenDfa();
    int m_statecount;
    vector< vector<FrozenDfaTransition> > m_transitions;

    int transition(int state, int c);
};

class LIBGLR_EXPORT NfaTransition
{
public:
    int m_from;
    int m_to;
    UIntSet m_on;
    NfaTransition() : m_from(-1), m_to(-1) {}
    NfaTransition(int from, int to) : m_from(from), m_to(to) {}
    bool operator==(const NfaTransition &rhs) const;
    bool operator<(const NfaTransition &rhs) const;
	static int NfaTransition_cmp(const NfaTransition *lhs, const NfaTransition *rhs);
};

class LIBGLR_EXPORT NfaETransition
{
public:
    int m_from;
    int m_to;
    NfaETransition() : m_from(-1), m_to(-1) {}
    NfaETransition(int from, int to) : m_from(from), m_to(to) {}
    bool operator==(const NfaETransition &rhs) const;
    bool operator<(const NfaETransition &rhs) const;
	static int NfaETransition_cmp(const NfaETransition *lhs, const NfaETransition *rhs);
};

class LIBGLR_EXPORT NfaDiagramHelper
{
public:
	NfaDiagramHelper() {}
	virtual ~NfaDiagramHelper() {}
	virtual String NodeLabel(Nfa &nfa, int node) = 0;
	virtual String TransitionLabel(Nfa &nfa, NfaTransition &trns) = 0;
	virtual String ETransitionLabel(Nfa &nfa, NfaETransition &etrns) = 0;
};

class EdgeSet
{
public:
	UIntSet m_inputs;
	UIntSet m_nextset;
};

class LIBGLR_EXPORT Nfa
{
    int m_nextstate;
public:
    vector<NfaTransition> m_transitions;
    bool m_TransitionsNeedSort;
    vector<NfaETransition> m_etransitions;
    bool m_eTransitionsNeedSort;
    vector<int> m_markers;
    vector<int> m_startstates;
    vector<int> m_endstates;

    int AddRxRec(Rx *rx, int initstate);
    void SortTransitions();

    Nfa();
    int addstate();
    int statecount();
    void addstartstate(int state);
    void addendstate(int state);
    void markstate(int state, int marker);
    void relocatedst(int olddst, int newdst);
    void addtransition_empty(int from, int to);
    void addtransition_char(int from, int to, int on);
    void addtransition_charset(int from, int to, const UIntSet &iset_on);
    void closure(vector<int> &seed);
    void discover_edges(const vector<int> &from, vector<EdgeSet> &edges);
    void transitionfnct(int input, int from, vector<int> &to);

    void AddLiteral(const char *str, bool case_insensitive);
    void AddRx(Rx *rx);
    void AddNfa(Nfa &src);
    void AsDfa(Nfa &dfa, vector< vector<int> > &substates);
    void Freeze(FrozenDfa &dfa);

	void Print(outbuf *s, NfaDiagramHelper *pHelper, String name);
};

#endif /* __nfa_h */
