#include <stdio.h>
#ifdef _WIN32
#include <windows.h>
typedef HANDLE handle_t;
#define H_STDIN GetStdHandle(STD_INPUT_HANDLE)
#define H_STDOUT GetStdHandle(STD_OUTPUT_HANDLE)
#define H_STDERR GetStdHandle(STD_ERROR_HANDLE)
typedef __int64 foff_t;
#else
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
typedef int handle_t;
#define H_STDIN STDIN_FILENO
#define H_STDOUT STDOUT_FILENO
#define H_STDERR STDERR_FILENO
typedef long long foff_t;
#endif

#define BUF_BLOCK_SIZE (32767)
#define SZ(a) (sizeof(a)/sizeof(a[0]))

#ifdef __cplusplus
extern "C" {
#endif

#define INBUF_SET (0)
#define INBUF_CUR (1)
#define INBUF_END (2)

struct tag_inbuf;
typedef struct tag_inbuf inbuf;
typedef int (*ibf_read)(inbuf*,void*,int);
typedef foff_t (*ibf_seek)(inbuf*,foff_t,int);
typedef void (*ibf_close)(inbuf*);

GLR_API inbuf *new_inbuf(size_t inbufsize, const char *fname, ibf_read readfnct, ibf_seek seekfnct, ibf_close closefnct);
GLR_API void free_inbuf(inbuf *buf);
GLR_API int ib_read(inbuf *buf, void *vp, int count);
GLR_API foff_t ib_seek(inbuf *buf, foff_t to, int whence);
GLR_API foff_t ib_tell(inbuf *buf);

struct tag_inbuf
{
  char *s;
  const char *fname;
  foff_t startpos;
  int count;
  int capacity;
  int pos;
  int eof;
  int line;
  int col;
  int strbuf;
  ibf_read readfnct;
  ibf_seek seekfnct;
  ibf_close closefnct;
#ifdef __cplusplus
  int read(void *vp,int len) { return ib_read(this,vp,len); }
  foff_t seek(foff_t off, int whence) { return ib_seek(this,off,whence); }
  foff_t tell() { return ib_tell(this); }
  void close() { if(!closefnct) return; closefnct(this); }
  void free() { free_inbuf(this); }
#endif
};

struct tag_fpinbuf; typedef struct tag_fpinbuf fpinbuf;
struct tag_fpinbuf {
    inbuf base;
    handle_t h;
};

GLR_API inbuf *h_inbuf(const char *fname, handle_t h, int handoff);
GLR_API inbuf *file_inbuf(const char *fname);
GLR_API inbuf *string_inbuf(const char *s, int staticstr);

struct tag_outbuf;
typedef struct tag_outbuf outbuf;
typedef void (*obf_write)(outbuf *buf, const char *s, size_t len);
typedef void (*obf_flush)(outbuf *buf);
typedef void (*obf_close)(outbuf *buf);

GLR_API outbuf *new_ob(size_t size, obf_write write, obf_flush flush, obf_close close);
GLR_API void free_ob(outbuf *buf);
GLR_API void ob_write(outbuf *buf, const char *s, size_t len);
GLR_API void ob_printf(outbuf *buf, const char *fmt, ...);
GLR_API void ob_vprintf(outbuf *buf, const char *fmt, va_list args);
GLR_API void ob_flush(outbuf *buf);
GLR_API void ob_putc(outbuf *buf, int c);
GLR_API void ob_puts(outbuf *buf, const char *s);

struct tag_outbuf
{
    char buffer[BUF_BLOCK_SIZE];
    char safetynet[16];
    size_t bufused;
    int enc;
    obf_write writefnct;
    obf_flush flushfnct;
    obf_close closefnct;
    int lineno;
    int colno;
    int autoflush;
#ifdef __cplusplus
    void write(const char *s, size_t len) { ob_write(this,s,len); }
    void printf(const char *fmt, ...)
    {
        va_list args;
        va_start(args,fmt);
        ob_vprintf(this,fmt,args);
        va_end(args);
    }
    void vprintf(const char *fmt, va_list args) { ob_vprintf(this,fmt,args); }
    void flush(outbuf *buf) { ob_flush(this); }
    void putC(int c) { ob_putc(this,c); }
    void puts(const char *s) { ob_puts(this,s); }
    void free() { free_ob(this); }
#endif
};

typedef struct tag_file_outbuf
{
    outbuf base;
    handle_t h;
} file_outbuf;

GLR_API outbuf *new_file_outbuf(const char *fname);
GLR_API outbuf *new_h_outbuf(const char *fname, handle_t h, int handoff, int autoflush);

typedef struct tag_string_outbuf
{
    outbuf base;
    int capacity, count;
    char *buf;
} string_outbuf;

GLR_API outbuf *new_string_outbuf();
GLR_API void obs_clear(outbuf *buf);
GLR_API const char *obs_getstr(outbuf *buf);

#ifdef __cplusplus
} /* extern "C" */
#endif
