#ifdef __cplusplus
extern "C" {
#endif
ttype ptr_tt = {"*", sizeof(void*), 0, 0};

GLR_API int lxstack_push(lxstack *stk, int i)
{
  int newcapacity = 0;
  int *newi = 0;
  if( stk->count == stk->capacity ) {
    newcapacity = stk->capacity ? (stk->capacity * 2) : 16;
    newi = (int*)vp_realloc(stk->i, newcapacity*sizeof(int));
    if( ! newi ) return 0;
    stk->i = newi;
    stk->capacity = newcapacity;
  }
  stk->i[stk->count++] = i;
  return 1;
}
GLR_API int lxstack_pop(lxstack *stk)
{
  if( stk->count == 0 ) return -1;
  return stk->i[--stk->count];
}
GLR_API int lxstack_size(lxstack *stk) { return stk->count; }
GLR_API void lxstack_clear(lxstack *stk)
{
  if( stk->i ) { vp_realloc(stk->i, 0); stk->i = 0; }
  stk->count = 0;
  stk->capacity = 0;
}
GLR_API int lxstack_top(lxstack *stk)
{
  if( ! stk->count ) return -1;
  return stk->i[stk->count-1];
}
GLR_API toknode *new_toknode(size_t size, ttype *tt, int sym, int state)
{
  toknode *node = (toknode*)vp_realloc(0, sizeof(toknode));
  vp_memzero(node, sizeof(toknode));
  if( size ) {
    node->vp = vp_realloc(0, size);
    vp_memzero(node->vp, size);
    node->size = size;
    node->tt = tt;
    if( tt && tt->init )
        tt->init(node->vp);
  }
  node->ref = 1;
  node->sym = sym;
  node->state = state;
  return node;
}
GLR_API void toknode_addref(toknode *node) { if(node) ++node->ref; }
GLR_API void toknode_destroytt(toknode *node)
{
  if( ! node || ! node->vp )
    return;
  if( node->tt && node->tt->destroy )
    node->tt->destroy(node->vp);
  vp_realloc(node->vp, 0);
  node->size = 0;
  node->vp = 0;
  node->tt = 0;
}
GLR_API void toknode_flatten(toknode *node)
{
    size_t i = 0, j = 0, endj = 0;
    toknode *cursubnode = 0;
    if( node->wstop )
    {
        subnodes_add(node, node->wstop);
        node->wstop = 0;
    }
    if( node->wsbottom )
    {
        subnodes_add(node, node->wsbottom);
        node->wsbottom = 0;
    }
    // notice that node->sncount may grow on each iteration.
    for( i = 0; i < node->sncount; ++i )
    {
        cursubnode = node->subnodes[i];
        if( ! cursubnode || cursubnode->ref > 1 )
            continue;
        if( cursubnode->wstop )
        {
            subnodes_add(node, cursubnode->wstop);
            cursubnode->wstop = 0;
        }
        if( cursubnode->wsbottom )
        {
            subnodes_add(node, cursubnode->wsbottom);
            cursubnode->wsbottom = 0;
        }
        j = 0;
        endj = cursubnode->sncount;
        for( ; j < endj; ++j )
            subnodes_add(node, cursubnode->subnodes[j]);
        cursubnode->sncount = 0;
    }
}
GLR_API void toknode_decref(toknode *node)
{
  size_t i;
  if( ! node )
    return;
  if( --node->ref > 0 )
    return;
  toknode_destroytt(node);
  if(node->tokstr) {
    vp_realloc(node->tokstr, 0);
    node->tokstr = 0;
  }
  /*
  The original code looked like this...

  if(node->wstop) {
    toknode_decref(node->wstop);
    node->wstop = 0;
  }
  if(node->wsbottom) {
    toknode_decref(node->wsbottom);
    node->wsbottom = 0;
  }
  for( i = 0; i < node->sncount; ++i )
    toknode_decref(node->subnodes[i]);

  unfortunately, we would get stack overflow
  when the call tree was very deep.  So instead,
  we'll decref by iteration instead of recursion.
  */
  if( node->wstop || node->wsbottom || node->sncount )
  {
    toknode_flatten(node);
    for( i = 0; i < node->sncount; ++i )
        toknode_decref(node->subnodes[i]);
  }
  node->sncount = node->sncapacity = 0;
  vp_realloc(node->subnodes, 0);
  node->subnodes = 0;
  vp_realloc(node, 0);
}
GLR_API int typestack_bump_capacity(typestack *tstack) {
  size_t newcapacity = 0;
  toknode **nodes = 0;
  if( tstack->tncount == tstack->tncapacity )
  {
    newcapacity = tstack->tncapacity ? (tstack->tncapacity * 2) : 16;
    nodes = (toknode**)vp_realloc(tstack->nodes, newcapacity*sizeof(toknode*));
    if( ! nodes )
      return 0;
    tstack->nodes = nodes;
    tstack->tncapacity = newcapacity;
  }
  return 1;
}
GLR_API void *typestack_alloc(typestack *tstack, ttype *tt, int sym, int state) {
  toknode *node = 0;
  size_t size = tt ? tt->size : 0;
  if( ! typestack_bump_capacity(tstack) )
      return 0;
  node = new_toknode(size, tt, sym, state);
  tstack->nodes[tstack->tncount++] = node;
  return node->vp;
}
GLR_API size_t measure_str(toknode *node, int ws) {
  size_t i, slen = 0;
  if( ! node )
    return 0;
  if( ws && node->wstop )
    slen += measure_str(node->wstop,1);
  if( node->sncount == 0 && node->tokstr )
      slen += node->tokslen;
  else {
    for( i = 0; i < node->sncount; ++i )
      slen += measure_str(node->subnodes[i],1);
  }
  if( ws && node->wsbottom )
    slen += measure_str(node->wsbottom,1);
  return slen;
}
GLR_API char *collect_str(toknode *node, char *dst, int ws) {
  size_t i = 0, n = 0;
  if( ! node )
    return dst;
  if( ws && node->wstop )
    dst = collect_str(node->wstop, dst, 1);
  if( node->sncount == 0 && node->tokstr ) {
    for( i = 0, n = node->tokslen; i < n; ++i )
      dst[i] = node->tokstr[i];
    dst += n;
  } else {
    for( i = 0; i < node->sncount; ++i )
      dst = collect_str(node->subnodes[i], dst, 1);
  }
  if( ws && node->wsbottom )
    dst = collect_str(node->wsbottom, dst, 1);
  return dst;
}
GLR_API char *toknode_str(toknode *node) {
  if( ! node )
    return "";
  if( node->sncount == 0 && node->tokstr )
    return node->tokstr;
  node->tokslen = measure_str(node,0);
  node->tokstr = (char*)vp_realloc(node->tokstr,node->tokslen+1);
  *collect_str(node,node->tokstr,0) = 0;
  return node->tokstr;
}
GLR_API int typestack_pop(typestack *stk, size_t count)
{
  int i = 0, n = 0;
  if( count > stk->tncount )
    return 0;
  for( i = (int)stk->tncount-1, n = (int)(stk->tncount-count); i >= n; --i )
    toknode_decref(stk->nodes[i]);
  stk->tncount -= count;
  return 1;
}
GLR_API int typestack_moveone(typestack *dststk, typestack *srcstk)
{
  if( srcstk->tncount == 0 )
    return 0;
  if( ! typestack_bump_capacity(dststk) )
    return 0;
  dststk->nodes[dststk->tncount++] = srcstk->nodes[--srcstk->tncount];
  return 1;
}
GLR_API void typestack_clear(typestack *stk)
{
  typestack_pop(stk, stk->tncount);
  vp_realloc(stk->nodes, 0);
  stk->nodes = 0;
  stk->tncount = 0;
  stk->tncapacity = 0;
}
GLR_API size_t typestack_size(typestack *stk) { return stk->tncount; }
GLR_API toknode *typestack_top(typestack *stk)
{
  if( ! stk->tncount ) return 0;
  return stk->nodes[stk->tncount-1];
}
GLR_API int subnodes_bump_capacity(toknode *node) {
  size_t newcapacity = 0;
  toknode **subnodes = 0;
  if( node->sncount == node->sncapacity )
  {
    newcapacity = node->sncapacity ? (node->sncapacity * 2) : 16;
    subnodes = (toknode**)vp_realloc(node->subnodes, newcapacity*sizeof(toknode*));
    if( ! subnodes )
      return 0;
    node->subnodes = subnodes;
    node->sncapacity = newcapacity;
  }
  return 1;
}
GLR_API int subnodes_add(toknode *node, toknode *sub) {
  if( ! subnodes_bump_capacity(node) ) return 0;
  node->subnodes[node->sncount++] = sub;
  return 1;
}
GLR_API toknode *makews(toknode *wsdst, toknode *wssrc) {
    toknode *tmp = 0;
    if( ! wsdst ) return wssrc;
    if( ! wssrc ) return wsdst;
    if( wsdst->sncount ) {
      subnodes_add(wsdst, wssrc);
      return wsdst; }
    tmp = new_toknode(0,0,0,0);
    subnodes_add(tmp, wsdst);
    subnodes_add(tmp, wssrc);
    return tmp;
}
GLR_API void toknode_set_subnodes(toknode *node, toknode **nodes, int sncount) {
  int i = 0;
  node->subnodes = (toknode**)vp_realloc(node->subnodes, sizeof(toknode*)*sncount);
  node->sncount = node->sncapacity = sncount;
  for( ; i < sncount; ++i )
    node->subnodes[i] = nodes[i];
  // copy leading and trailing ws to the nonterminal
  for( i = 0; i < sncount && node->subnodes[i] == 0; ++i );
  if( i < sncount && node->subnodes[i]->wstop ) {
    node->wstop = node->subnodes[i]->wstop;
    node->subnodes[i]->wstop = 0;
  }
  for( i = sncount-1; i >= 0 && node->subnodes[i] == 0; --i );
  if( i >= 0 && node->subnodes[i]->wsbottom ) {
    node->wsbottom = node->subnodes[i]->wsbottom;
    node->subnodes[i]->wsbottom = 0;
  }
}
GLR_API void toknode_unset_subnodes(toknode *node, toknode **nodes, int sncount) {
  int i = 0;
  // copy leading and trailing ws back to the terminals
  for( i = 0; i < sncount && node->subnodes[i] == 0; ++i );
  if( i < sncount && node->subnodes[i]->wstop ) {
    node->subnodes[i]->wstop = node->wstop;
    node->wstop = 0;
  }
  for( i = sncount-1; i >= 0 && node->subnodes[i] == 0; --i );
  if( i >= 0 && node->subnodes[i]->wsbottom ) {
    node->subnodes[i]->wsbottom = node->wsbottom;
    node->wsbottom = 0;
  }
  node->sncount = node->sncapacity = 0;
  (toknode**)vp_realloc(node->subnodes, 0);
  node->subnodes = 0;
}
GLR_API void toknode_keep_subnodes(toknode *node) {
  size_t i = 0, sncount = node->sncount;
  for( ; i < sncount; ++i )
    toknode_destroytt(node->subnodes[i]);
}
GLR_API void update_buf_line_col(inbuf *buf, int len)
{
  int i;
  for( i = 0; i < len; ++i ) {
      if( buf->s[buf->pos+i] == '\n' ) { buf->line++; buf->col = 1; } else buf->col++;
  }
}
#ifdef NO_COMPRESSED_SCANNER
GLR_API int next_state(int c, int curs, int lexs, int *plastaccepttok, tsinfo *tsinfos)
{
    lxstnfo *lxinfo = lxinfo = tsinfos[lexs].lxinfos+curs;
    int low = 0, high = lxinfo->icount-1, mid = 0, nexts = -1;
    lxtr *ranges = 0;
    *plastaccepttok = -1;
    while( low <= high ) {
        mid = (low+high)/2; ranges = lxinfo->toranges + mid;
        if( c < ranges->low ) { high = mid-1; continue; }
        if( c > ranges->high ) { low = mid+1; continue; }
        nexts = ranges->next;
        if( ranges->tok != -1 ) *plastaccepttok = ranges->tok;
        break;
    }
    return nexts;
}
#else
GLR_API int next_state(int c, int curs, int lexs, int *plastaccepttok, tsinfo *tsinfos)
{
    tsinfo *tsfo = tsinfos+lexs;
    int brk = -1, high, low, mid, seek, s;
    scnitem *scnit = 0;

    low = 0; high = tsfo->brkcount-2;
    while(low<=high)
    {
        mid = (low+high)/2;
        if( c < tsfo->brks[mid] )
        {
            high = mid-1;
            continue;
        }
        if( c >= tsfo->brks[mid+1] )
        {
            low = mid+1;
            continue;
        }
        break;
    }
    brk = mid;

    high = tsfo->scncnt-1;
    seek = curs*tsfo->brkcount+brk;
    while(1)
    {
        low = 0;
        while( low <= high )
        {
            mid = (low+high)/2;
            scnit = tsfo->scnitems+mid;
            if( seek < scnit->pos )
            {
                high = mid-1;
                continue;
            }
            if( seek > scnit->pos+scnit->count )
            {
                low = mid+1;
                continue;
            }
            break;
        }
        if(low>high)
            return -1;
        if(scnit->pos+scnit->count == seek)
            break;
        high = mid-1;
        seek = scnit->off + ((seek - scnit->pos) % scnit->lead);
    }
    s = scnit->next;
    if(s < 0 || s >= tsfo->statecount)
    {
        s = -1;
        *plastaccepttok = -1;
    }
    else
        *plastaccepttok = tsfo->accepttoks[s];
    return s;
}
#endif
GLR_API int readtok(inbuf *buf, lxstack *lexstk, typestack *instk, outbuf *trace, int *isws, outbuf *oberr, int eof, tsinfo *tsinfos, symbol *allsymbols, lxact **lxactions)
{
  int pos = 0, s = 0, lexs = 0, c = 0, accepttok = -1, lastaccepttok = -1, lastacceptlen = 0, i = 0, n = 0, low = 0, high = 0, mid = 0, rd = 0, nc = 0, fr = 0, mbdec_i = 0, mbdec_n = 0;
  char **ptok = 0, *tok = 0, *tmps = 0;
  toknode *node = 0;
  lxact *tkact = 0;
  lexs = lxstack_top(lexstk);
  pos = buf->pos;
  *isws = 0;
  while( c != -1 && s != -1 ) {
    if( pos == buf->count ) {
      //read more buffer if you can
      if( ! buf->eof ) {
        if( buf->pos >= BUF_BLOCK_SIZE ) {
            //use the space in the existing buffer
            fr = (buf->pos / BUF_BLOCK_SIZE) * BUF_BLOCK_SIZE;
            buf->startpos += fr;
            for( i = 0, n = buf->capacity-fr; i < n; ++i ) buf->s[i] = buf->s[fr+i];
            buf->count -= fr;
            buf->pos -= fr;
            pos -= fr;
        }
        else {
            //make more space in the buffer
            nc = buf->capacity + BUF_BLOCK_SIZE; fr = BUF_BLOCK_SIZE;
            tmps = (char*)vp_realloc(buf->s, nc);
            if( ! tmps ) return 0;
            buf->s = tmps; buf->capacity = nc;
        }
        rd = buf->readfnct(buf, buf->s+buf->count, fr);
        if( rd < fr )
          buf->eof = 1;
        buf->count += rd;
      }
      //we are really out of buffer
      if( pos == buf->count ) {
        //we're out, but we've consumed some buffer, so it's a token, or an error.
        if( pos > buf->pos )
          break;
        //we're out, and haven't consumed any buffer, so it's EOF
        typestack_alloc(instk, 0, eof, -1);
        node = typestack_top(instk);
        node->line = buf->line;
        node->col = buf->col;
        return 1;
      }
    }
    //get the next code point
    if( ! mbdec_n )
    {
        c = ((unsigned char*)buf->s)[pos++];
        if( c > 0x7F )
        {
            // started a multi-byte fetch.
            mbdec_i = 0;
            mbdec_n = 0;
            while( c & 0x0080 )
            {
                c <<= 1;
                mbdec_n++;
            }
            c &= 0x007F;
            c >>= mbdec_n;
        }
    }
    // if we've started or are continuing a multi-byte fetch, proceed.
    while(mbdec_n && pos < buf->count)
    {
        --mbdec_n;
        rd = ((unsigned char*)buf->s)[pos++];
	    if( (rd & 0xC0000000) != 0x80000000 )
        {
            update_buf_line_col(buf, pos-buf->pos);
            ob_printf(oberr, "%s:%d:%d: error: tokenizer stall, bad multibyte encoding byte %d\n", buf->fname, buf->line, buf->col, mbdec_i+2);
		    return 0;
        }
	    if( mbdec_i == 0 && (rd & 0x3FFFFFFF) == 0 )
        {
            update_buf_line_col(buf, pos-buf->pos);
            ob_printf(oberr, "%s:%d:%d: error: tokenizer stall, overlong encoding\n", buf->fname, buf->line, buf->col);
		    return 0;
        }
        c = (c<<6)|(rd&0x7F);
        mbdec_i++;
    }
    if( mbdec_n )
    {
        if( ! buf->eof )
            continue;
        update_buf_line_col(buf, pos-buf->pos);
        ob_printf(oberr, "%s:%d:%d: error: tokenizer stall, eof while doing multi-byte decoding\n", buf->fname, buf->line, buf->col);
	    return 0;        
    }
    //do a binary search for the next state
    s = next_state(c, s, lexs, &accepttok, tsinfos);
    if( accepttok != -1 )
    {
        lastaccepttok = accepttok;
        lastacceptlen = pos-buf->pos;
    }
  }
  if( lastaccepttok == -1 ) {
      ob_printf(oberr, "%s:%d:%d: error: tokenizer stall\n", buf->fname, buf->line, buf->col); return 0; }
  if( trace )
    ob_printf(trace, "read token %s\n", allsymbols[lastaccepttok].name);
  ptok = (char**)typestack_alloc(instk, &ptr_tt, lastaccepttok, -1);
  node = typestack_top(instk);
  node->line = buf->line;
  node->col = buf->col;
  node->tokslen = lastacceptlen;
  node->tokstr = tok = *ptok = (char*)vp_realloc(0, lastacceptlen+1);
  for( i = 0; i < lastacceptlen; ++i ) tok[i] = buf->s[buf->pos+i];
  tok[lastacceptlen] = 0;
  update_buf_line_col(buf, lastacceptlen);
  buf->pos += lastacceptlen;
  tkact = lxactions[lexs];
  while( tkact->sym != -1 && tkact->sym != lastaccepttok )
    ++tkact;
  switch(tkact->flags){
    case PUSH_FLAG: lxstack_push(lexstk, tkact->next); break;
    case POP_FLAG:
      lxstack_pop(lexstk);
      if( lexstk->count == 0 ) {
        ob_printf(oberr, "empty lexer state stack\n"); return 0; }
      break;
    case GOTO_FLAG: lexstk->i[lexstk->count-1] = tkact->next; break;
    case WS_FLAG:
      *isws = 1;
      break;
    default: break;
  }
  return 1;
}

#ifdef NO_COMPRESSED_LR_TABLE
#define ACTION_SYMBOL(a,s) ((a)->shiftto_or_reduceby)
#else
#define FIRST_SYMBOL(a) ((a)->range[1])
#define LAST_SYMBOL(a) ((a)->range[(a)->range[0]*2])
#define ACTION_SYMBOL(a,s) (((a)->first)+(((a)->first==(a)->last)?0:((s)-FIRST_SYMBOL(a))))
#endif

GLR_API int update_parse_state(typestack *prsstk, typestack *instk, toknode **nodes, inbuf *buf, outbuf *trace, line_col *err_lc,
                                parse_state *parse_states,
                                production *productions,
                                int production_adjust_count,
                                symbol *allsymbols,
                                int starttok,
                                int errtok,
                                int acceptstate,
                                outbuf *oberr,
                                reduce_action_fnct ra,
                                void *pextra
                                )
{
  int state = -1, tok = -1, low = 0, high = 0, mid = 0, lows = 0, highs = 0, gotlc = 0, stkpos = -1;
  parse_state *ps = 0;
#ifdef NO_COMPRESSED_LR_TABLE
  parse_action *action = 0, *lastaction = 0, *endaction = 0;
#else
  parse_action_range *action = 0, *lastaction = 0, *endaction = 0;
#endif
  production *p = 0, *pO = 0; ttype *tt = 0;
#ifndef NO_COMPRESSED_LR_TABLE
  int *srange = 0;
#endif
  int *skips = 0, *endskips = 0;
  toknode **curnode = 0, **dstnode = 0, *ntnode = 0, *errnode = 0;
  
  while( typestack_size(instk) ) {
    tok = typestack_top(instk)->sym;
    state = typestack_top(prsstk)->state;
    if( state < 0 ) break;
    ps = parse_states+state;
    low = 0; high = ps->acount-1;
    while( low <= high ) {
      mid = (low+high)/2;
      action = ps->actions+mid;
#ifdef NO_COMPRESSED_LR_TABLE
      if( tok < action->token ) { high = mid-1; continue; }
      if( tok > action->token ) { low = mid+1; continue; }
#else
      if( tok < FIRST_SYMBOL(action) ) { high = mid-1; continue; }
      if( tok > LAST_SYMBOL(action) ) { low = mid+1; continue; }
#endif
      break;
    }
#ifdef NO_COMPRESSED_LR_TABLE
    if(tok != action->token) {
#else
    if( low <= high ) {
        // we found a candidate action range, now see if it pans out.
        low = 0; high = action->range[0]-1;
        while( low <= high ) {
          mid = (low+high)/2;
          srange = action->range+1+mid*2;
          if( tok < srange[0] ) { high = mid-1; continue; }
          if( tok > srange[1] ) { low = mid+1; continue; }
          break;
        }
    }
    if( low > high ) {
#endif
parse_error:
      // An error!
      if( typestack_top(prsstk)->sym == errtok )
      {
          if(trace)
            ob_printf(trace, "handling parse error: discarding token\n", tok);
          typestack_pop(instk, 1);
          continue;
      }
      // Search for an error handling production
      for(stkpos=(int)typestack_size(prsstk)-1;stkpos>=0;--stkpos)
      {
          tok = prsstk->nodes[stkpos]->sym;
          state = prsstk->nodes[stkpos]->state;
          if(state<0) { stkpos=-1; break; }
          ps = parse_states+state;
          if(!ps->acount) continue;
          mid = 0;
#if NO_COMPRESSED_LR_TABLE
          while(ps->actions[mid].token<errtok) ++mid;
          if( ps->actions[mid].token == errtok && ps->actions[mid].action == SH )
#else
          while( LAST_SYMBOL(ps->actions+mid) < errtok) ++mid;
          if( errtok >= FIRST_SYMBOL(ps->actions+mid) && errtok <= LAST_SYMBOL(ps->actions+mid) && ps->actions[mid].action == SH )
#endif
            break; // found it, stkpos != -1
      }
      if(stkpos==-1)
      {
        // No production we are building handles errors, stop parsing.
        ntnode = typestack_top(instk);
        ob_printf(oberr, "%s:%d:%d: error: parse error\n", buf->fname, ntnode->line, ntnode->col);
        if( err_lc ) {
          err_lc->line = ntnode->line;
          err_lc->col = ntnode->col;
        }
        return 0;
      }
      // This production handles errors, shift the error and carry on.
      action = ps->actions+mid;
      typestack_pop(prsstk, stkpos-typestack_size(prsstk)-1);
      if(trace)
        ob_printf(trace, "parse error: shift to %d\n", ACTION_SYMBOL(action,errtok));
      typestack_alloc(prsstk, 0, errtok, ACTION_SYMBOL(action,errtok));
      errnode = typestack_top(prsstk);
      ntnode = typestack_top(instk);
      errnode->line = ntnode->line;
      errnode->col = ntnode->col;
      continue;
    }
    if( action->action == SH ) {
      if( trace )
        ob_printf(trace, "shift to %d\n", ACTION_SYMBOL(action,tok));
      if( ! typestack_moveone(prsstk, instk) )
        return 0;
      typestack_top(prsstk)->state = ACTION_SYMBOL(action,tok);
      continue;
    }
#ifdef NO_COMPRESSED_LR_TABLE
    // might be multiple reductions available for resolvable conflicts
    while( action > ps->actions && action[-1].token == tok )
      --action;
#endif
    // find the last action.  It is usually the very next one.
    endaction = ps->actions+ps->acount;
    for( lastaction = action+1; lastaction < endaction; ++lastaction )
    {
#ifdef NO_COMPRESSED_LR_TABLE
      if( tok != lastaction->token )
        break;
#else
      if( tok < FIRST_SYMBOL(action) || tok > LAST_SYMBOL(action) )
        break;
#endif
    }
    // Use the first reduction that isn't rejected
    for( ; action!=lastaction; ++action ) {
      p = productions+ACTION_SYMBOL(action,tok);
      pO = productions+(p->origid-production_adjust_count);
      skips = p->skips; endskips = skips+p->skcount;
      tt = allsymbols[p->nt].tt;
      typestack_alloc(instk, tt, pO->id, -1);
      ntnode = typestack_top(instk);
      dstnode = nodes;
      gotlc = 0;
      *dstnode++ = ntnode;
      curnode = prsstk->nodes+(prsstk->tncount-p->scount);
      for( lows = 0, highs = pO->scount; lows < highs; ++lows )
        if( skips < endskips && lows == *skips )
            *dstnode++ = 0, ++skips;
        else {
            if( ! gotlc ) {
              ntnode->line = (*curnode)->line; ntnode->col = (*curnode)->col; gotlc = 1; }
            *dstnode++ = *curnode++;
        }
      if( ! gotlc ) { ntnode->line = buf->line; ntnode->col = buf->col; }
      toknode_set_subnodes(ntnode, nodes+1, pO->scount);
      if( trace )
        ob_printf(trace, "reduce %d by production %d", p->scount, p->id-production_adjust_count);
      if( ra(nodes, p->origid, buf->fname, ntnode->line, ntnode->col, oberr, pextra) ) {
        if( trace )
          ob_printf(trace, " (ok)\n");
        toknode_keep_subnodes(ntnode);
        prsstk->tncount -= p->scount;
        if( p->nt == starttok ) {
          if( ! typestack_moveone(prsstk, instk) )
            return 0;
          typestack_top(prsstk)->state = acceptstate;
          return 1;
        }
        break;
      }
      if( trace )
        ob_printf(trace, " (failed)\n");
      toknode_unset_subnodes(ntnode, nodes+1, pO->scount);
      typestack_pop(instk, 1);
    }
    if( action == lastaction )
	goto parse_error;
  }
  return 1;
}

GLR_API int parse_any(inbuf *buf, outbuf *obuf, outbuf *trace, line_col *err_lc,
                                tsinfo *tsinfos,
                                lxact **lxactions,
                                parse_state *parse_states,
                                production *productions,
                                int production_adjust_count,
                                int max_production_symbols,
                                symbol *allsymbols,
                                int starttok,
                                int errtok,
                                int eoftok,
                                int acceptstate,
                                int parser_start_state,
                                int initial_tokenset,
                                reduce_action_fnct ra,
                                void *pextra)
{
  size_t i, n;
  lxstack lexstk = NULL_LXSTACK;
  typestack instk = NULL_TYPESTACK;
  typestack prsstk = NULL_TYPESTACK;
  int ret = 1, isws = 0;
  toknode **nodes = (toknode**)vp_realloc(0, (max_production_symbols+1)*sizeof(toknode*));
  if( err_lc ) { err_lc->line = 0; err_lc->col = 0; }
  lxstack_push(&lexstk, initial_tokenset);
  typestack_alloc(&prsstk, &ptr_tt, 0, parser_start_state);
  do {
    if( ! readtok(buf, &lexstk, &instk, trace, &isws, obuf, eoftok, tsinfos, allsymbols, lxactions) ) {
      if( err_lc ) {
        err_lc->line = buf->line;
        err_lc->col = buf->col;
      }
      ret = 0; goto cleanup; }
    if( isws ) {
      //append whitespace to the end of the last token
      prsstk.nodes[prsstk.tncount-1]->wsbottom = makews(prsstk.nodes[prsstk.tncount-1]->wsbottom, instk.nodes[--instk.tncount]);
      continue; }
    //move accumulated whitespace from the end of the last token to the beginning of the next one
    if( prsstk.nodes[prsstk.tncount-1]->wsbottom ) {
      instk.nodes[instk.tncount-1]->wstop = makews(instk.nodes[instk.tncount-1]->wstop, prsstk.nodes[prsstk.tncount-1]->wsbottom);
      prsstk.nodes[prsstk.tncount-1]->wsbottom = 0;
    }
    if( ! update_parse_state(&prsstk, &instk, nodes, buf, trace,
                                        err_lc, parse_states, productions, production_adjust_count,
                                        allsymbols, starttok, errtok, acceptstate, obuf,
                                        ra, pextra) ) {
        ret = 0; goto cleanup; }
    if( trace ) {
      ob_printf(obuf, "state stack:");
      for( i = 0, n = prsstk.tncount; i < n; ++i )
        ob_printf(obuf, " %d", prsstk.nodes[i]->state);
      ob_printf(obuf, "\n");
    }
  } while( typestack_top(&prsstk)->state != acceptstate );
cleanup:
  vp_realloc(nodes, 0);
  typestack_clear(&prsstk); typestack_clear(&instk); lxstack_clear(&lexstk);
  return ret;
}

#ifdef __cplusplus
} /* extern "C" */
#endif
