#define NULL_TYPESTACK {0, 0, 0}
#define NULL_LXSTACK {0, 0, 0}

#define WS_FLAG   (1)
#define PUSH_FLAG (2)
#define POP_FLAG  (3)
#define GOTO_FLAG (4)

#define STKVALUE(stk,t,i) (*((t*)((stk)->nodes[(stk)->tncount+(i)].vp)))
#define SH (0)
#define RD (1)

#ifdef __cplusplus
#include <new>
template<class T> void init_T(void *vp) { new(vp) T; }
template<class T> void destruct_T(void *t) { ((T*)t)->~T(); }
#define INIT_T(T)     init_T<T>
#define DESTRUCT_T(T) destruct_T<T>
#else
#define INIT_T(T)     0
#define DESTRUCT_T(T) 0
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef struct tag_lxstack { int *i, count, capacity; } lxstack;
GLR_API int lxstack_push(lxstack *stk, int i);
GLR_API int lxstack_pop(lxstack *stk);
GLR_API int lxstack_size(lxstack *stk);
GLR_API void lxstack_clear(lxstack *stk);
GLR_API int lxstack_top(lxstack *stk);

typedef struct tag_ttype {
  const char *name;
  size_t size;
  void (*init)(void *vp);
  void (*destroy)(void *vp);
} ttype;

extern ttype ptr_tt;

struct tag_toknode;
typedef struct tag_toknode toknode;
struct tag_toknode {
  int ref;
  int line, col;
  int sym, state;
  void *vp;
  size_t size;
  ttype *tt;
  size_t tokslen; char *tokstr;
  toknode *wstop, *wsbottom;
  size_t sncount, sncapacity; toknode **subnodes;
};

GLR_API toknode *new_toknode(size_t size, ttype *tt, int sym, int state);
GLR_API void toknode_addref(toknode *node);
GLR_API void toknode_destroytt(toknode *node);
GLR_API void toknode_decref(toknode *node);

typedef struct tag_typestack
{
  toknode **nodes;
  size_t tncount, tncapacity;
} typestack;

GLR_API int typestack_bump_capacity(typestack *tstack);
GLR_API void *typestack_alloc(typestack *tstack, ttype *tt, int sym, int state);
GLR_API size_t measure_str(toknode *node, int ws);
GLR_API char *collect_str(toknode *node, char *dst, int ws);
GLR_API char *toknode_str(toknode *node);
GLR_API int typestack_pop(typestack *stk, size_t count);
GLR_API int typestack_moveone(typestack *dststk, typestack *srcstk);
GLR_API void typestack_clear(typestack *stk);
GLR_API size_t typestack_size(typestack *stk);
GLR_API toknode *typestack_top(typestack *stk);

typedef struct tag_symbol { const char *name; ttype *tt; } symbol;
typedef struct tag_production { int id, origid, nt, scount, *symbols, skcount, *skips; } production;
#ifdef NO_COMPRESSED_LR_TABLE
typedef struct tag_parse_action { int token, action, shiftto_or_reduceby; } parse_action;
typedef struct tag_parse_state { int acount; parse_action *actions; } parse_state;
#else
typedef struct tag_parse_action_range { int action, first, last, *range; } parse_action_range;
typedef struct tag_parse_state { int acount; parse_action_range *actions; } parse_state;
#endif
#ifdef NO_COMPRESSED_SCANNER
typedef struct tag_lxtr { int low, high, next, tok; } lxtr;
typedef struct tag_lxstnfo { int from, icount; lxtr *toranges; } lxstnfo;
typedef struct tag_tsinfo { int tokstate, lxcount; lxstnfo *lxinfos; } tsinfo;
#else
typedef struct tag_scnitem { int pos, off, count, lead, next; } scnitem;
typedef struct tag_tsinfo { int brkcount, *brks, statecount, *accepttoks, scncnt; scnitem *scnitems; } tsinfo;
#endif
GLR_API int subnodes_bump_capacity(toknode *node);
GLR_API int subnodes_add(toknode *node, toknode *sub);
GLR_API toknode *makews(toknode *wsdst, toknode *wssrc);
GLR_API void toknode_set_subnodes(toknode *node, toknode **nodes, int sncount);
GLR_API void toknode_unset_subnodes(toknode *node, toknode **nodes, int sncount);
GLR_API void toknode_keep_subnodes(toknode *node);

typedef struct tag_line_col {
  int line, col;
} line_col;

typedef struct lxact_tag { int sym, flags, next; } lxact;

GLR_API void update_buf_line_col(inbuf *buf, int len);

GLR_API int readtok(inbuf *buf, lxstack *lexstk, typestack *instk, outbuf *trace, int *isws, outbuf *oberr, int eof, tsinfo *tsinfos, symbol *allsymbols, lxact **lxactions);

typedef int (*reduce_action_fnct)(toknode **nodes, int pid, const char *fname, int line, int col, outbuf *oberr, void *pextra);

GLR_API int update_parse_state(typestack *prsstk, typestack *instk, toknode **nodes, inbuf *buf, outbuf *trace, line_col *err_lc,
                                parse_state *parse_states,
                                production *productions,
                                int production_adjust_count,
                                symbol *allsymbols,
                                int starttok,
                                int errtok,
                                int acceptstate,
                                outbuf *oberr,
                                reduce_action_fnct ra,
                                void *pextra
                                );

GLR_API int parse_any(inbuf *buf, outbuf *obuf, outbuf *trace, line_col *err_lc,
                                tsinfo *tsinfos,
                                lxact **lxactions,
                                parse_state *parse_states,
                                production *productions,
                                int production_adjust_count,
                                int max_production_symbols,
                                symbol *allsymbols,
                                int starttok,
                                int errtok,
                                int eoftok,
                                int acceptstate,
                                int parser_start_state,
                                int initial_tokenset,
                                reduce_action_fnct ra,
                                void *pextra);

#ifdef __cplusplus
} /* extern "C" */
#endif
