// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __sortedset_h
#define __sortedset_h

#include "vec.h"

template<class T>
class sortedset
{
private:
    vector<T> m_vec;
    int (*m_cmp)(const T *lhs, const T *rhs);

public:
    typedef T *iterator;
    typedef const T *const_iterator;

    sortedset(int (*cmp)(const T *lhs, const T *rhs) = 0) : m_cmp(cmp)
    {
        if( ! m_cmp )
             m_cmp = simplecmp<T>;
    }

    bool operator==(const sortedset<T> &rhs) const
    {
        return m_vec == rhs.m_vec;
    }

	iterator begin()
	{
		return m_vec.begin();
	}

	iterator end()
	{
		return m_vec.end();
	}

	const_iterator begin() const
	{
		return m_vec.begin();
	}

	const_iterator end() const
	{
		return m_vec.end();
	}

    const_iterator find(const T &rhs) const
    {
        typename vector<T>::iterator first, last;
        if( ! m_vec.binsearch(rhs, first, last, m_cmp) )
            return m_vec.end();
        return first;
    }
    
    iterator find(const T &rhs)
    {
        typename vector<T>::iterator first, last;
        if( ! m_vec.binsearch(rhs, first, last, m_cmp) )
            return m_vec.end();
        return m_vec.begin()+(first-m_vec.begin());
    }

    iterator insert(const T &rhs)
    {
        typename vector<T>::iterator first, last;
        if( m_vec.binsearch(rhs, first, last, m_cmp) )
            return m_vec.end();
        return m_vec.insertat(first-m_vec.begin(), rhs);
    }

    void insert(const_iterator first, const_iterator last)
    {
        while( first != last )
        {
            insert(*first);
            ++first;
        }
    }

    void erase(const T &t)
    {
        typename vector<T>::iterator first, last;
        if( ! m_vec.binsearch(t, first, last, m_cmp) )
            return;
        if( first != last )
            m_vec.erase(first, last);
    }

    void erase(iterator first, iterator last)
    {
        m_vec.erase(first, last);
    }

    iterator erase(iterator iter)
    {
        return m_vec.erase(iter);
    }

    size_t size() const
    {
        return m_vec.size();
    }

    void clear()
    {
        m_vec.clear();
    }
};

#endif /* __sortedset_h */
