// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __tokensetmerge_h
#define __tokensetmerge_h

#include "grammar.h"

class LIBGLR_EXPORT MergeTokenSetStatesException
{
};

class AlteredProductionInState;

class LIBGLR_EXPORT ProductionInState
{
public:
    Production *m_p;
    OrderedTokenSet *m_state;
    Production *m_clone; // Does not participate in comparison operators
    AlteredProductionInState *m_altered_production; // Does not participate in comparison operators

public:
    ProductionInState();
    ProductionInState(Production *p, OrderedTokenSet *state);
    bool operator<(const ProductionInState &rhs) const;
    bool operator==(const ProductionInState &rhs) const;
    bool operator!=(const ProductionInState &rhs) const { return ! (*this == rhs); }
};

class LIBGLR_EXPORT SymbolInState
{
public:
    Symbol *m_s;
    OrderedTokenSet *m_state;
    Symbol *m_clone; // Does not participate in comparison operators
    TokenSetActions *m_aggregate_nt_actions; // Does not participate in comparison operators

public:
    SymbolInState();
    SymbolInState(Symbol *s, OrderedTokenSet *state);
    bool operator<(const SymbolInState &rhs) const;
    bool operator==(const SymbolInState &rhs) const;
    bool operator!=(const SymbolInState &rhs) const { return ! (*this == rhs); }
};

class LIBGLR_EXPORT AlteredProductionInState
{
public:
    AlteredProductionInState(Production *p, OrderedTokenSet *state);
    Production *m_p;
    OrderedTokenSet *m_state;
    vector<SymbolInState> m_items;
};

class LIBGLR_EXPORT TokenSetMiss
{
public:
    TokenSetMiss();
    TokenSetMiss(SymbolInState s, ProductionInState p, int pos);
    SymbolInState m_s;
    ProductionInState m_p;
    int m_pos;

    bool operator==(const TokenSetMiss &rhs) const;
    bool operator<(const TokenSetMiss &rhs) const;

    void Print(outbuf *pio);
};

class LIBGLR_EXPORT TokenSetMergeAlgorithm
{
protected:
    vector<Production*> m_productionordering;
    vector<SymbolInState> m_clones;
    vector<SymbolInState> m_nondeterministic_tokens;
    vector<ProductionInState> m_productions;
    bool IsSymbolCloneable(Symbol *s, OrderedTokenSet *tokset, bool bForceYes);
    void FinalizeActions(Symbol *s, OrderedTokenSet *tokset, TokenSetActions &finalactions);
    bool GetFinalActions(Symbol *s, OrderedTokenSet *tokset, TokenSetActions* &aggregate_nt_actions);
    AlteredProductionInState *GetProductionInState(Production *p, OrderedTokenSet *tokset);
    void RecordTokenSetMiss(SymbolInState s, ProductionInState p, int pos);
    void Symbol_PreMergeTokenSetStatesRec(Symbol *s, OrderedTokenSet *tokset, vector<OrderedTokenSet*> &statestack, TokenSetActions &finalactions);
    void Production_PreMergeTokenSetStatesRec(Production *p, OrderedTokenSet *tokset, vector<OrderedTokenSet*> &statestack, TokenSetActions &finalactions);
    Symbol *CloneSymbol(Symbol *orig, OrderedTokenSet *ts);
    Production *CloneProduction(Production *porig, Symbol *clonent, AlteredProductionInState *altered_production);

    outbuf *m_err;

public:
    Grammar *m_g;
    sortedset<TokenSetMiss> m_TokenSetMiss;

    TokenSetMergeAlgorithm(Grammar *g, outbuf *err);
    ~TokenSetMergeAlgorithm();
    void Collect();
    void Apply();
    void Report(outbuf *pio);
};

#endif /* __tokensetmerge_h */
